/*************************************************************************/
/* Copyright (C) 1995-1997 Marlis Hochbruck, Christian Lubich, Hubert    */
/* Selhofer. All rights reserved.                                        */
/*                                                                       */
/* C code written by Mathias Froehlich.                                  */
/*                                                                       */
/* This code is part of a copyrighted package. For details, see the file */
/* "COPYING" in the top-level directory.                                 */
/*************************************************************************/

#ifndef CGS_H
#define CGS_H

#ifdef  __cplusplus
extern "C" {
#endif

typedef struct {
  int NExactBreakdown;
  int ReCalculateV;
  int VStride;
  finteger Size;
  Complex *C;
  Complex *V;
  Complex *S;
  Complex *Sp;
  Complex *R;
  Complex *Rp;
  Complex D;
  Complex Dp;
  Complex Nu;
  Complex Mu;
  int C_Eq_V0;

  int LastValidV;
} CgsPrivat;

extern void PREC(Exp4FillHandleWithCgs) (struct Krylov *Kry,
      finteger MaxKrylovDim,finteger neq,int HaveOwnH);

/* accesses the size if the V base */
#define CgsSize(Data)            (KrylovPrivat(Data).Cgs.Size)

#define CgsReCalculateV(Data)    (KrylovPrivat(Data).Cgs.ReCalculateV)

#define CgsC_Eq_V0(Data)        (KrylovPrivat(Data).Cgs.C_Eq_V0)

/* accesses the start vector of the left base */
#define CgsC(Data)               (KrylovPrivat(Data).Cgs.C)

/* accesses the pointer to V */
#define CgsVData(Data)           (KrylovPrivat(Data).Cgs.V)

/* accesses the i - th entry of V */
#define CgsV(Data,i)           (KrylovPrivat(Data).Cgs.V+(i)*KrylovLdBig(Data))

/* accesses the pointer to S */
#define CgsS(Data)               (KrylovPrivat(Data).Cgs.S)

/* accesses the pointer to Sp */
#define CgsSp(Data)              (KrylovPrivat(Data).Cgs.Sp)

/* accesses the pointer to R */
#define CgsR(Data)               (KrylovPrivat(Data).Cgs.R)

/* accesses the pointer to Rp */
#define CgsRp(Data)              (KrylovPrivat(Data).Cgs.Rp)

#define CgsIncrementExactBreakdownCount(Data) \
((KrylovPrivat(Data).Cgs.NExactBreakdown)++)

#define CgsClearExactBreakdownCount(Data) \
((KrylovPrivat(Data).Cgs.NExactBreakdown)=0)

#define CgsExactBreakdownCount(Data) (KrylovPrivat(Data).Cgs.NExactBreakdown)

#define CgsD(Data)               (KrylovPrivat(Data).Cgs.D)

#define CgsDp(Data)              (KrylovPrivat(Data).Cgs.Dp)

#define CgsNu(Data)              (KrylovPrivat(Data).Cgs.Nu)

#define CgsMu(Data)              (KrylovPrivat(Data).Cgs.Mu)

#define CgsVStride(Data)         (KrylovPrivat(Data).Cgs.VStride)

#define CgsLastValidV(Data)      (KrylovPrivat(Data).Cgs.LastValidV)

#ifdef  __cplusplus
}
#endif
#endif
