/*************************************************************************/
/* Copyright (C) 1995-1997 Marlis Hochbruck, Christian Lubich, Hubert    */
/* Selhofer. All rights reserved.                                        */
/*                                                                       */
/* C code written by Mathias Froehlich.                                  */
/*                                                                       */
/* This code is part of a copyrighted package. For details, see the file */
/* "COPYING" in the top-level directory.                                 */
/*************************************************************************/

/* header for errors.c */

#ifndef EXP4_ERRORS_H
#define EXP4_ERRORS_H

#include <stdlib.h>
#include <stdio.h>
#include <setjmp.h>

#define ErrorReturn(comment,val) {					      \
fprintf(stderr,"Exp4 Error, occured in source file " __FILE__ 		      \
      " at line %d:\n" comment "\n",__LINE__);				      \
Exp4LongJumpExit (val);							      \
}

#define ReturnMessage(comment,ret) {					      \
fprintf(stderr,"Exp4 Error, occured in source file " __FILE__		      \
      " at line %d:\n" comment "\n",__LINE__);				      \
      return ret;							      \
}

#define Warning(comment) fprintf(stderr,comment "\n")

#ifdef SINGLE_PRECISION
#define lapack_err(Info,SubRoutine,LaRoutine)				      \
if (Info != 0) {							      \
  fprintf(stderr,"Exp4 Error: in subroutine S" SubRoutine " s" LaRoutine      \
	" returns with info = %d\n",(int)Info);				      \
  Exp4LongJumpExit (EXP4_ERR_LAPACK);					      \
}
#else
#ifdef DOUBLE_PRECISION
#define lapack_err(Info,SubRoutine,LaRoutine)				      \
if (Info != 0) {							      \
  fprintf(stderr,"Exp4 Error: in subroutine D" SubRoutine " d" LaRoutine      \
	" returns with info = %d\n",(int)Info);				      \
  Exp4LongJumpExit (EXP4_ERR_LAPACK);					      \
}
#else
#ifdef SINGLE_COMPLEX
#define lapack_err(Info,SubRoutine,LaRoutine)				      \
if (Info != 0) {							      \
  fprintf(stderr,"Exp4 Error: in subroutine C" SubRoutine " c" LaRoutine      \
	" returns with info = %d\n",(int)Info);				      \
  Exp4LongJumpExit (EXP4_ERR_LAPACK);					      \
}
#else
#ifdef DOUBLE_COMPLEX
#define lapack_err(Info,SubRoutine,LaRoutine)				      \
if (Info != 0) {							      \
  fprintf(stderr,"Exp4 Error: in subroutine Z" SubRoutine " z" LaRoutine      \
	" returns with info = %d\n",(int)Info);				      \
  Exp4LongJumpExit (EXP4_ERR_LAPACK);					      \
}
#else
#define lapack_err(Info,SubRoutine,LaRoutine)				      \
if (Info != 0) {							      \
  fprintf(stderr,"Exp4 Error: in subroutine " SubRoutine " " LaRoutine	      \
	" returns with info = %d\n",(int)Info);				      \
  Exp4LongJumpExit (EXP4_ERR_LAPACK);					      \
}
#endif
#endif
#endif
#endif

/* define error Codes */
#define EXP4_DETECT_SINGULARITY                                         1
#define EXP4_ERR_NO_MEM                                                -1
#define EXP4_ERR_OUTPUT                                                -7
#define EXP4_ERR_NO_RHS                                                -8
#define EXP4_ERR_NO_JAC                                                -9
#define EXP4_ERR_NO_MV                                                -10
#define EXP4_ERR_NO_VM                                                -11
#define EXP4_ERR_INVALID_KRY_TYPE                                     -12
#define EXP4_ERR_LAPACK                                               -14
#define EXP4_ERR_NO_DENSE                                             -15
#define EXP4_ERR_EXACT_BREAKDOWN                                      -16
#define EXP4_ERR_IN_WORKSPACE_DISTRIBUTION                            -17
#define EXP4_ERR_TRY_TO_REBUILD_UNKNOWN_KRYLOV_SPACE                  -19

#ifdef  __cplusplus
extern "C" {
#endif

extern void Exp4SetJumpBuf(jmp_buf *in);

extern void Exp4LongJumpExit (int ret);

#ifdef  __cplusplus
}
#endif

#endif
