/*************************************************************************/
/* Copyright (C) 1995-1997 Marlis Hochbruck, Christian Lubich, Hubert    */
/* Selhofer. All rights reserved.                                        */
/*                                                                       */
/* C code written by Mathias Froehlich.                                  */
/*                                                                       */
/* This code is part of a copyrighted package. For details, see the file */
/* "COPYING" in the top-level directory.                                 */
/*************************************************************************/

#ifndef LANCZOS_H
#define LANCZOS_H

#ifdef  __cplusplus
extern "C" {
#endif

typedef struct {
  finteger MaxBlockSize;
  Complex *V;
  Complex *W;
  Complex D;
  Complex *BetaTemp;
  Complex *BetaMult;
  Real Norm1BetaV,Norm1BetaW;
  finteger *RegularIndex;
  int RegularIndexIndex;
  finteger LastValidV;
  int NExactBreakdown;
  int WStride;
  int VStride;
} LanczosPrivat;

/* accesses the pointer to V */
#define LanczosVData(Data)     (KrylovPrivat(Data).Lanczos.V)

/* accesses the i - th entry of V */
#define LanczosV(Data,i) \
(KrylovPrivat(Data).Lanczos.V+(i)*KrylovLdBig(Data))

/* accesses the pointer to W */
#define LanczosWData(Data)     (KrylovPrivat(Data).Lanczos.W)

/* accesses the i - th entry of W */
#define LanczosW(Data,i) \
(KrylovPrivat(Data).Lanczos.W+(i)*KrylovLdBig(Data))

/* the periode for overwriting the base vectors */
#define LanczosVStride(Data)   (KrylovPrivat(Data).Lanczos.VStride)
#define LanczosWStride(Data)   (KrylovPrivat(Data).Lanczos.WStride)

/* macros to count the number of exact breakdowns */
#define LanczosIncrementExactBreakdownCount(Data) \
((KrylovPrivat(Data).Lanczos.NExactBreakdown)++)

#define LanczosClearExactBreakdownCount(Data) \
((KrylovPrivat(Data).Lanczos.NExactBreakdown)=0)

#define LanczosExactBreakdownCount(Data) \
(KrylovPrivat(Data).Lanczos.NExactBreakdown)

/* acess the diagonal entrie of the matrix d for the last block */
#define LanczosLastD(Data)      (KrylovPrivat(Data).Lanczos.D)

/* holds the last V base vector which is not overwritten by an new one */
#define LanczosLastValidV(Data) (KrylovPrivat(Data).Lanczos.LastValidV)

/* used to restore the Lanczos base exactly */
#define LanczosBetaMult(Data)    (KrylovPrivat(Data).Lanczos.BetaMult)
#define LanczosBetaTemp(Data)    (KrylovPrivat(Data).Lanczos.BetaTemp)

/* holds the 1-norms of the linear combination coeficients
 * for the temporary vector */
#define LanczosNorm1BetaV(Data)  (KrylovPrivat(Data).Lanczos.Norm1BetaV)
#define LanczosNorm1BetaW(Data)  (KrylovPrivat(Data).Lanczos.Norm1BetaW)

/* holds the maximum block size for the Lanczos process */
#define LanczosMaxBlockSize(Data) (KrylovPrivat(Data).Lanczos.MaxBlockSize)

/* acess the regular indices */
#define LanczosReadRegularIndex(Data,i) \
(KrylovPrivat(Data).Lanczos.RegularIndex[max((i),0)])

#define LanczosRegularIndexData(Data) (KrylovPrivat(Data).Lanczos.RegularIndex)

#define LanczosRegularIndexRelativ(Data,Rel) \
(KrylovPrivat(Data).Lanczos.RegularIndex[ \
max(KrylovPrivat(Data).Lanczos.RegularIndexIndex+(Rel),0)])

#define LanczosSetRegularIndex(Data,I) \
((KrylovPrivat(Data).Lanczos.RegularIndex[ \
++(KrylovPrivat(Data).Lanczos.RegularIndexIndex)])=(I))

#define LanczosClearRegularIndexArray(Data) \
(KrylovPrivat(Data).Lanczos.RegularIndexIndex = -1, \
      LanczosSetRegularIndex(Data,0))
    

extern void PREC(Exp4FillHandleWithLanczos) (struct Krylov *Kry,
      finteger MaxKrylovDim,finteger MaxBlockSize,finteger neq,int HaveOwnH);

#ifdef  __cplusplus
}
#endif
#endif
