/*************************************************************************/
/* Copyright (C) 1995-1997 Marlis Hochbruck, Christian Lubich, Hubert    */
/* Selhofer. All rights reserved.                                        */
/*                                                                       */
/* C code written by Mathias Froehlich.                                  */
/*                                                                       */
/* This code is part of a copyrighted package. For details, see the file */
/* "COPYING" in the top-level directory.                                 */
/*************************************************************************/

#include <stdlib.h>
#include <stdio.h>
#include <math.h>
#include <limits.h>
#include "ftypes.h"
#include "fblas.h"
#include "prec.h"

#include "phi.h"

#define PADE_N       6

/* the pade coeff's */
static Real pade_n[PADE_N]={1.0/26.0,5.0/156.0,1.0/858.0,
				     1.0/5720.0,1.0/205920.0,1.0/8648640.0};
static Real pade_d[PADE_N]={-6.0/13.0,5.0/52.0,-5.0/429.0,1.0/1144.0,
				     -1.0/25740.0,1.0/1235520.0};

Complex PREC(phi) (Complex z)
{
  if (pabs(z) < 1.0/2.0)
  {
#if (defined(SINGLE_PRECISION) || defined(DOUBLE_PRECISION))
    Complex e = 1.0 + z/26.0;
    Complex d = 1.0 - z*6.0/13.0;
    Complex x = z;
    finteger i;

    for (i=1;i<PADE_N;i++)
    {
      x *= z;
      e += pade_n[i] * x;
      d += pade_d[i] * x;
    }

    return e/d;
#else
    Complex e,d,x,res;
    Real help;
    finteger i;

    e.r = 1.0 + z.r/26.0;
    e.i =       z.i/26.0;

    d.r = 1.0 - z.r*6.0/13.0;
    d.i =     - z.i*6.0/13.0;

    x.r = z.r;
    x.i = z.i;

    for (i=1;i<PADE_N;i++)
    {
      help = x.r*z.r - x.i*z.i;
      x.i  = x.i*z.r + x.r*z.i;
      x.r  = help;

      e.r += pade_n[i]*x.r;
      e.i += pade_n[i]*x.i;

      d.r += pade_d[i]*x.r;
      d.i += pade_d[i]*x.i;
    }

    /* result=x=e/d */
    help = d.r*d.r + d.i*d.i;
    res.r = (e.r*d.r + e.i*d.i)/help;
    res.i = (e.i*d.r - e.r*d.i)/help;

    return res;
#endif
  }
  else
  {
#if (defined(SINGLE_PRECISION) || defined(DOUBLE_PRECISION))
/* #define HAVE_EXPM1 */
#ifdef HAVE_EXPM1
    return expm1(z)/z;
#else
    return (exp(z)-1.0)/z;
#endif
#else
    Complex res;
    Real help=cos(z.i);

    res.r=z.r*help;  /* = re(z)*cos(im(z)) */
    res.i=-z.i*help; /* = -im(z)*cos(im(z)) */

    help=sin(z.i);
    res.r+=z.i*help; /* = re(z)*cos(im(z)) + im(z)*sin(im(z)) */
    res.i+=z.r*help; /* = re(z)*sin(im(z)) - im(z)*cos(im(z)) */

    help=exp(z.r);
    res.r*=help;/* = exp(re(z)) * (re(z)*cos(im(z)) + im(z)*sin(im(z))) */
    res.i*=help;/* = exp(re(z)) * (re(z)*sin(im(z)) - im(z)*cos(im(z))) */


    res.r-=z.r;
    /* = exp(re(z)) * (re(z)*cos(im(z)) + im(z)*sin(im(z))) - re(z) */

    res.i+=z.i;
    /* = exp(re(z)) * (re(z)*sin(im(z)) - im(z)*cos(im(z))) + im(z) */

    help = z.r*z.r+z.i*z.i;
    res.r/=help;/* = (exp(re(z)) *
		    (re(z)*cos(im(z)) + im(z)*sin(im(z))) - re(z))/(|z|*|z|) */

    res.i/=help;/* = (exp(re(z)) *
		    (re(z)*sin(im(z)) - im(z)*cos(im(z))) + im(z))/(|z|*|z|) */

    return res;
#endif
  }
}
