/*************************************************************************/
/* Copyright (C) 1995-1997 Marlis Hochbruck, Christian Lubich, Hubert    */
/* Selhofer. All rights reserved.                                        */
/*                                                                       */
/* C code written by Mathias Froehlich.                                  */
/*                                                                       */
/* This code is part of a copyrighted package. For details, see the file */
/* "COPYING" in the top-level directory.                                 */
/*************************************************************************/


#ifndef PREC_H
#define PREC_H

/* default make only DOUBLE_PRECISION */
#if (!defined(SINGLE_PRECISION) && !defined(DOUBLE_PRECISION) \
     && !defined(SINGLE_COMPLEX) && !defined(DOUBLE_COMPLEX))
#error "You have to define a Precision !"
#endif

#ifdef SINGLE_PRECISION
#define REAL
#define SINGLE
#endif
#ifdef DOUBLE_PRECISION
#define REAL
#define DOUBLE
#endif
#ifdef SINGLE_COMPLEX
#define COMPLEX
#define SINGLE
#endif
#ifdef DOUBLE_COMPLEX
#define COMPLEX
#define DOUBLE
#endif

/*
 * Two stage defines for concatenation.
 */
#define __EXP4_CAT__(a,b,c,d) a##b##c##d
#define CAT(a,b,c,d) __EXP4_CAT__(a,b,c,d)

/*
 * Two stage defines for stringification.
 */
#define __EXP4_STRING__(a) #a
#define STRING(a) __EXP4_STRING__(a)

#ifdef SINGLE
# define _RPREFIX s
# define _CPREFIX c
#elif defined(DOUBLE)
# define _RPREFIX d
# define _CPREFIX z
#else
# error "None of SINGLE and DOUBLE is defined!"
#endif

#ifdef REAL
# define RPREFIX
# define CPREFIX RPREFIX
#elif defined(COMPLEX)
# define RPREFIX _RPREFIX
# define CPREFIX _CPREFIX
#else
# error "None of REAL and COMPLEX is defined!"
#endif

#if 0 /* FIXME */
#ifdef HAVE_NATIVE_CBLAS
# include "cblas.h"

# define paxpy CAT(cblas_,CPREFIX,axpy,)
# define pscal CAT(cblas_,CPREFIX,scal,)
# define pswap CAT(cblas_,CPREFIX,swap,)
# define pcopy CAT(cblas_,CPREFIX,copy,)
# define pnrm2 CAT(cblas_,RPREFIX,CPREFIX,nrm2)
# define pasum CAT(cblas_,RPREFIX,CPREFIX,asum)

#else

# define paxpy(N, ALPHA, DX, IX, DY, IY)				      \
{									      \
  const finteger _n = (N);						      \
  const finteger _ix = (IX);						      \
  const finteger _iy = (IY);						      \
  const Complex _alpha = (ALPHA);					      \
  F77CALL (CAT(CPREFIX,axpy,,)) (&_n, &_alpha, DX, &_ix, DY, &_iy);	      \
}

# define pscal(N, ALPHA, DX, IX)					      \
{									      \
  const finteger _n = (N);						      \
  const finteger _ix = (IX);						      \
  const Complex _alpha = (ALPHA);					      \
  F77CALL (CAT(CPREFIX,scal,,)) (&_n, &_alpha, DX, &_ix);		      \
}

# define pswap(N, DX, IX, DY, IY)					      \
{									      \
  const finteger _n = (N);						      \
  const finteger _ix = (IX);						      \
  const finteger _iy = (IY);						      \
  F77CALL (CAT(CPREFIX,swap,,)) (&_n, DX, &_ix, DY, &_iy);		      \
}

# define pcopy(N, DX, IX, DY, IY)					      \
{									      \
  const finteger _n = (N);						      \
  const finteger _ix = (IX);						      \
  const finteger _iy = (IY);						      \
  F77CALL (CAT(CPREFIX,copy,,)) (&_n, DX, &_ix, DY, &_iy);		      \
}

# define pnrm2(N, DX, IX)						      \
{									      \
  const finteger _n = (N);						      \
  const finteger _ix = (IX);						      \
  F77CALL (CAT(CPREFIX,RPREFIX,nrm2,)) (&_n, DX, &_ix);			      \
}

# define pasum(N, DX, IX)						      \
{									      \
  const finteger _n = (N);						      \
  const finteger _ix = (IX);						      \
  F77CALL (CAT(CPREFIX,RPREFIX,asum,)) (&_n, DX, &_ix);			      \
}



/* level 1 blas */
#define prot   F77CALL (srot)
#define pdot(res,n,x,incx,y,incy) (*(res)) = F77CALL (sdot) (n,x,incx,y,incy)
#define prscl  F77CALL (srscl)

/* level 2 blas */
#define pgemv  F77CALL (sgemv)
#define prgemv  F77CALL (sgemv)
#define ptrsv  F77CALL (strsv)
#define pger   F77CALL (sger)
#define pgbmv  F77CALL (sgbmv)

/* level 3 blas */
#define pgemm  F77CALL (sgemm)
#define ptrsm  F77CALL (strsm)


#endif


#define prscal F77CALL (sscal)
#define praxpy  F77CALL (saxpy)
#endif



/* use single precision */
#ifdef SINGLE_PRECISION

typedef fsingle Complex;
#define ComplexConst(value)      (value)
#define ComplexSetReal(var,value) (var)=(value)
#define ComplexIsZero(Val)       ((Val) == 0.0)

typedef fsingle Real;

typedef fsingle RealCoef;
#define RealCoefConst(value)      (value)
#define SetRealCoefVal(var,value) (var)=(value)

#define PREC(name) S##name

#define Re(var) (var)
#define Im(var) 0.0
#define radd(res,r,c)  (res) = (r) + (c)
#define add(res,a,b)   (res) = (a) + (b)
#define sub(res,a,b)   (res) = (a) - (b)
#define rmul(res,r,c)  (res) = (r) * (c)
#define mul(res,r,c)   (res) = (r) * (c)
#define neg(res,c)     (res) =-(c)
#define conj(res,c)    (res) = (c)
#define pdiv(res,a,b)   (res) = (a)/(b)
#define rdiv(res,a,b)  (res) = (a)/(b)
#define pabs(value)    fabs(value)
#define rabs(value)    fabs(value)
#define ZERO(value)    ((value) == 0.0)

/* level 1 blas */
#define paxpy  F77CALL (saxpy)
#define praxpy  F77CALL (saxpy)
#define pswap  F77CALL (sswap)
#define pcopy  F77CALL (scopy)
#define pscal  F77CALL (sscal)
#define prscal F77CALL (sscal)
#define pnrm2  F77CALL (snrm2)
#define pasum  F77CALL (sasum)
#define prot   F77CALL (srot)
#define pdot(res,n,x,incx,y,incy) (*(res)) = F77CALL (sdot) (n,x,incx,y,incy)
#define prscl  F77CALL (srscl)

/* level 2 blas */
#define pgemv  F77CALL (sgemv)
#define prgemv  F77CALL (sgemv)
#define ptrsv  F77CALL (strsv)
#define pger   F77CALL (sger)
#define pgbmv  F77CALL (sgbmv)

/* level 3 blas */
#define pgemm  F77CALL (sgemm)
#define ptrsm  F77CALL (strsm)

/* lapack */
#define plamch F77CALL (slamch)
#define plarnv F77CALL (slarnv)
#define pgeqpf(m,n,a,lda,jpvt,tau,work,info) \
F77CALL (sgeqpf) (m,n,a,lda,jpvt,tau,work,info) 
#define pormqr F77CALL (sormqr)
#define porm2r F77CALL (sorm2r)
#define ptrtrs F77CALL (strtrs)
#define plapmt F77CALL (slapmt)
#define pgesv  F77CALL (sgesv)
#define pgesvd F77CALL (sgesvd)
#define pgbsv  F77CALL (sgbsv)
#define pgetrf F77CALL (sgetrf)
#define pgetrs F77CALL (sgetrs)
#define placpy F77CALL (slacpy)
#define plascl F77CALL (slascl)
#define pgeev  F77CALL (sgeev)
#define pgecon F77CALL (sgecon)
#define ptrcon F77CALL (strcon)
#define plange F77CALL (slange)
#define plangb F77CALL (slangb)



#define plaset F77CALL (slaset)
#define phseqr F77CALL (shseqr)
#define pgebal F77CALL (sgebal)
#define pgehrd F77CALL (sgehrd)
#define porghr F77CALL (sorghr)
#define ptrevc F77CALL (strevc)
#define pgebak F77CALL (sgebak)
#define ipamax F77CALL (isamax)
#define plapy2 F77CALL (slapy2)
#define plartg F77CALL (slartg)
#define ptrsna F77CALL (strsna)

#endif

/* use double precision */
#ifdef DOUBLE_PRECISION

typedef fdouble Complex;
#define ComplexConst(value)      (value)
#define ComplexSetReal(var,value) (var)=(value)
#define ComplexIsZero(Val)       ((Val) == 0.0)

typedef fdouble RealCoef;
#define RealCoefConst(value)      (value)
#define SetRealCoefVal(var,value) (var)=(value)

typedef fdouble Real;

#define PREC(name) D##name

#define Re(var) (var)
#define Im(var) 0.0
#define radd(res,r,c)  (res) = (r) + (c)
#define add(res,a,b)   (res) = (a) + (b)
#define sub(res,a,b)   (res) = (a) - (b)
#define rmul(res,r,c)  (res) = (r) * (c)
#define mul(res,r,c)   (res) = (r) * (c)
#define neg(res,c)     (res) =-(c)
#define conj(res,c)    (res) = (c)
#define pdiv(res,a,b)   (res) = (a)/(b)
#define rdiv(res,a,b)  (res) = (a)/(b)
#define pabs(value)    fabs(value)
#define rabs(value)    fabs(value)
#define ZERO(value)    ((value) == 0.0)

/* level 1 blas */
#define paxpy  F77CALL (daxpy)
#define praxpy  F77CALL (daxpy)
#define pswap  F77CALL (dswap)
#define pcopy  F77CALL (dcopy)
#define pscal  F77CALL (dscal)
#define prscal F77CALL (dscal)
#define pnrm2  F77CALL (dnrm2)
#define pasum  F77CALL (dasum)
#define prot   F77CALL (drot)
#define pdot(res,n,x,incx,y,incy) (*(res)) = F77CALL (ddot) (n,x,incx,y,incy)
#define prscl  F77CALL (drscl)

/* level 2 blas */
#define pgemv  F77CALL (dgemv)
#define prgemv  F77CALL (dgemv)
#define ptrsv  F77CALL (dtrsv)
#define pger   F77CALL (dger)
#define pgbmv  F77CALL (dgbmv)

/* level 3 blas */
#define pgemm  F77CALL (dgemm)
#define ptrsm  F77CALL (dtrsm)

/* lapack */
#define plamch F77CALL (dlamch)
#define plarnv F77CALL (dlarnv)
#define pgeqpf(m,n,a,lda,jpvt,tau,work,info) \
F77CALL (dgeqpf) (m,n,a,lda,jpvt,tau,work,info) 
#define pormqr F77CALL (dormqr)
#define porm2r F77CALL (dorm2r)
#define ptrtrs F77CALL (dtrtrs)
#define plapmt F77CALL (dlapmt)
#define pgesv  F77CALL (dgesv)
#define pgesvd F77CALL (dgesvd)
#define pgbsv  F77CALL (dgbsv)
#define pgetrf F77CALL (dgetrf)
#define pgetrs F77CALL (dgetrs)
#define placpy F77CALL (dlacpy)
#define plascl F77CALL (dlascl)
#define pgeev  F77CALL (dgeev)
#define pgecon F77CALL (dgecon)
#define ptrcon F77CALL (dtrcon)
#define plange F77CALL (dlange)
#define plangb F77CALL (dlangb)


#define plaset F77CALL (dlaset)
#define phseqr F77CALL (dhseqr)
#define pgebal F77CALL (dgebal)
#define pgehrd F77CALL (dgehrd)
#define porghr F77CALL (dorghr)
#define ptrevc F77CALL (dtrevc)
#define pgebak F77CALL (dgebak)
#define ipamax F77CALL (idamax)
#define plapy2 F77CALL (dlapy2)
#define plartg F77CALL (dlartg)
#define ptrsna F77CALL (dtrsna)

#endif


/* use single precision complex */
#ifdef SINGLE_COMPLEX

typedef fsinglecomplex Complex;
#define ComplexConst(value)      {(value), 0.0}
#define ComplexSetReal(var,value) (var).r=(value),(var).i=0.0
#define ComplexIsZero(Val)       (((Val).r == 0.0) && ((Val).i == 0.0))

typedef fsingle Real;

#define PREC(name) C##name

#define Re(var) (var).r
#define Im(var) (var).i
#define radd(res,r,c)  (res).r = (r).r + (c).r, (res).i = (c).i
#define add(res,a,b)   (res).r = (a).r + (b).r, (res).i = (a).i + (b).i
#define sub(res,a,b)   (res).r = (a).r - (b).r, (res).i = (a).i - (b).i
#define rmul(res,ra,c) (res).r=(ra)*(c).r,(res).i=(ra)*(c).i
#define mul(res,a,b)  { fsingle __tmp=(a).r*(b).r-(a).i*(b).i; \
  (res).i=(a).r*(b).i+(a).i*(b).r; \
  (res).r=__tmp; }
#define neg(res,c)     (res).r=-(c).r,(res).i=-(c).i
#define conj(res,c)    (res).r=(c).r,(res).i=-(c).i
#define rdiv(res,a,b)   (res).r=(a).r/(b),(res).i=(a).i/(b)
#define pdiv(res,a,b)  { fsingle __absbsqare = (b).r*(b).r+(b).i*(b).i; \
    fsingle __tmp=((a).r*(b).r+(a).i*(b).i)/__absbsqare; \
    (res).i=((a).i*(b).r-(a).r*(b).i)/__absbsqare; \
    (res).r=__tmp; }
#define pabs(value)     sqrt((value).r*(value).r+(value).i*(value).i)
#define rabs(value)     fabs(value)
#define ZERO(value)     (((value).r == 0.0) && ((value).i == 0.0))

/* level 1 blas */
#define paxpy  F77CALL (caxpy)
#define pswap  F77CALL (cswap)
#define pcopy  F77CALL (ccopy)
#define pscal  F77CALL (cscal)
#define prscal F77CALL (csscal)
#define pnrm2  F77CALL (scnrm2)
#define pasum  F77CALL (scasum)
#define prot   F77CALL (crot)
#if (defined(USE_F2C) || defined(RETURN_COMPLEX_WITH_POINTER))
#define pdot(res,n,x,incx,y,incy) F77CALL (cdotc) (res,n,x,incx,y,incy)
#else
#define pdot(res,n,x,incx,y,incy) (*(res)) = F77CALL (cdotc) (n,x,incx,y,incy)
#endif
#define prscl  F77CALL (csrscl)

/* level 2 blas */
#define pgemv  F77CALL (cgemv)
#define ptrsv  F77CALL (ctrsv)
#define pger   F77CALL (cgeru)
#define pgbmv  F77CALL (cgbmv)

/* level 3 blas */
#define pgemm  F77CALL (cgemm)
#define ptrsm  F77CALL (ctrsm)

/* lapack */
#define plamch F77CALL (slamch)
#define plarnv F77CALL (clarnv)
#define pgeqpf(m,n,a,lda,jpvt,tau,work,info) \
F77CALL (cgeqpf) (m,n,a,lda,jpvt,tau,work,(Real*)((work)+*(n)),info)
#define pormqr F77CALL (cunmqr)
#define porm2r F77CALL (cunm2r)
#define ptrtrs F77CALL (ctrtrs)
#define plapmt F77CALL (clapmt)
#define pgesv  F77CALL (cgesv)
#define pgesvd F77CALL (cgesvd)
#define pgbsv  F77CALL (cgbsv)
#define pgetrf F77CALL (cgetrf)
#define pgetrs F77CALL (cgetrs)
#define placpy F77CALL (clacpy)
#define plascl F77CALL (clascl)
#define pgeev  F77CALL (cgeev)
#define pgecon F77CALL (cgecon)
#define ptrcon F77CALL (ctrcon)
#define plange(Norm,M,N,A,LdA,Work) \
 F77CALL (clange) (Norm,M,N,A,LdA,(Real*)(Work))
#define plangb(Norm,N,Kl,Ku,A,LdA,Work) \
 F77CALL (clangb) (Norm,N,Kl,Ku,A,LdA,(Real*)(Work))


#define plaset F77CALL (claset)
#define phseqr F77CALL (chseqr)
#define pgebal F77CALL (cgebal)
#define pgehrd F77CALL (cgehrd)
#define porghr F77CALL (cunghr)
#define ptrevc F77CALL (ctrevc)
#define pgebak F77CALL (cgebak)
#define ipamax F77CALL (icamax)
#define plapy2 F77CALL (clapy2)
#define plartg F77CALL (clartg)
#define ptrsna F77CALL (ctrsna)


#ifdef USE_SECURE_COMPLEX_ARITMETIC

typedef fsinglecomplex RealCoef;
#define RealCoefConst(value) {(value), 0.0}
#define SetRealCoefVal(var,value) (var).r=(value),(var).i=0.0

#define praxpy  F77CALL (caxpy)
#define prgemv  F77CALL (cgemv)
#else

typedef fsingle RealCoef;
#define RealCoefConst(value) (value)
#define SetRealCoefVal(var,value) (var)=(value)

/* _axpy with a real coeficient alpha */
#define praxpy(n,a,x,incx,y,incy) \
{ \
  finteger _2n_ = 2*(*n); \
  F77CALL (saxpy) (&_2n_,a,(Real*)(x),incx,(Real*)(y),incy); \
}

/* _gemv with a real coeficient vector y */
#define prgemv(trans,m,n,alpha,a,lda,x,incx,beta,y,incy) \
{ \
  finteger _2m_ = 2*(*m),_2lda_ = 2*(*lda); \
  F77CALL (sgemv) (trans,&_2m_,n,alpha,(Real*)(a),&_2lda_, \
	x,incx,beta,(Real*)(y),incy) ; \
}

#endif

#endif


/* use double precision complex */
#ifdef DOUBLE_COMPLEX

typedef fdoublecomplex Complex;
#define ComplexConst(value)      {(value), 0.0}
#define ComplexSetReal(var,value) (var).r=(value),(var).i=0.0
#define ComplexIsZero(Val)       (((Val).r == 0.0) && ((Val).i == 0.0))

typedef fdouble Real;

#define PREC(name) Z##name

#define Re(var) (var).r
#define Im(var) (var).i
#define radd(res,r,c)  (res).r = (r).r + (c).r, (res).i = (c).i
#define add(res,a,b)   (res).r = (a).r + (b).r, (res).i = (a).i + (b).i
#define sub(res,a,b)   (res).r = (a).r - (b).r, (res).i = (a).i - (b).i
#define rmul(res,ra,c) (res).r=(ra)*(c).r,(res).i=(ra)*(c).i
#define mul(res,a,b)  { fdouble __tmp=(a).r*(b).r-(a).i*(b).i; \
  (res).i=(a).r*(b).i+(a).i*(b).r; \
  (res).r=__tmp; }
#define neg(res,c)     (res).r=-(c).r,(res).i=-(c).i
#define conj(res,c)    (res).r=(c).r,(res).i=-(c).i
#define rdiv(res,a,b)   (res).r=(a).r/(b),(res).i=(a).i/(b)
#define pdiv(res,a,b)  { fdouble __absbsqare = (b).r*(b).r+(b).i*(b).i; \
    fdouble __tmp=((a).r*(b).r+(a).i*(b).i)/__absbsqare; \
    (res).i=((a).i*(b).r-(a).r*(b).i)/__absbsqare; \
    (res).r=__tmp; }
#define pabs(value)     sqrt((value).r*(value).r+(value).i*(value).i)
#define rabs(value)     fabs(value)
#define ZERO(value)     (((value).r == 0.0) && ((value).i == 0.0))

/* level 1 blas */
#define paxpy  F77CALL (zaxpy)
#define pswap  F77CALL (zswap)
#define pcopy  F77CALL (zcopy)
#define pscal  F77CALL (zscal)
#define prscal F77CALL (zdscal)
#define pnrm2  F77CALL (dznrm2)
#define pasum  F77CALL (dzasum)
#define prot   F77CALL (zrot)
#if (defined(USE_F2C) || defined(RETURN_COMPLEX_WITH_POINTER))
#define pdot(res,n,x,incx,y,incy) F77CALL (zdotc) (res,n,x,incx,y,incy)
#else
#define pdot(res,n,x,incx,y,incy) (*(res)) = F77CALL (zdotc) (n,x,incx,y,incy)
#endif
#define prscl  F77CALL (zdrscl)

/* level 2 blas */
#define pgemv  F77CALL (zgemv)
#define ptrsv  F77CALL (ztrsv)
#define pger   F77CALL (zgeru)
#define pgbmv  F77CALL (zgbmv)

/* level 3 blas */
#define pgemm  F77CALL (zgemm)
#define ptrsm  F77CALL (ztrsm)

/* lapack */
#define plamch F77CALL (dlamch)
#define plarnv F77CALL (zlarnv)
#define pgeqpf(m,n,a,lda,jpvt,tau,work,info) \
F77CALL (zgeqpf) (m,n,a,lda,jpvt,tau,work,(Real*)((work)+*(n)),info)
#define pormqr F77CALL (zunmqr)
#define porm2r F77CALL (zunm2r)
#define ptrtrs F77CALL (ztrtrs)
#define plapmt F77CALL (zlapmt)
#define pgesv  F77CALL (zgesv)
#define pgesvd F77CALL (zgesvd)
#define pgbsv  F77CALL (zgbsv)
#define pgetrf F77CALL (zgetrf)
#define pgetrs F77CALL (zgetrs)
#define placpy F77CALL (zlacpy)
#define plascl F77CALL (zlascl)
#define pgeev  F77CALL (zgeev)
#define pgecon F77CALL (zgecon)
#define ptrcon F77CALL (ztrcon)
#define plange(Norm,M,N,A,LdA,Work) \
 F77CALL (zlange) (Norm,M,N,A,LdA,(Real*)(Work))
#define plangb(Norm,N,Kl,Ku,A,LdA,Work) \
 F77CALL (zlangb) (Norm,N,Kl,Ku,A,LdA,(Real*)(Work))


#define plaset F77CALL (zlaset)
#define phseqr F77CALL (zhseqr)
#define pgebal F77CALL (zgebal)
#define pgehrd F77CALL (zgehrd)
#define porghr F77CALL (zunghr)
#define ptrevc F77CALL (ztrevc)
#define pgebak F77CALL (zgebak)
#define ipamax F77CALL (izamax)
#define plapy2 F77CALL (zlapy2)
#define plartg F77CALL (zlartg)
#define ptrsna F77CALL (ztrsna)

#ifdef USE_SECURE_COMPLEX_ARITMETIC

typedef fdoublecomplex RealCoef;
#define RealCoefConst(value)      {(value), 0.0}
#define SetRealCoefVal(var,value) (var).r=(value),(var).i=0.0

#define praxpy  F77CALL (zaxpy)
#define prgemv  F77CALL (zgemv)
#else

typedef fdouble RealCoef;
#define RealCoefConst(value)      (value)
#define SetRealCoefVal(var,value) (var)=(value)

/* _axpy with a real coeficient alpha */
#define praxpy(n,a,x,incx,y,incy) \
{ \
  const finteger _2n_ = 2*(*n); \
  F77CALL (daxpy) (&_2n_,a,(Real*)(x),incx,(Real*)(y),incy); \
}

/* _gemv with a real coeficient vector y */
#define prgemv(trans,m,n,alpha,a,lda,x,incx,beta,y,incy) \
{ \
  finteger _2m_ = 2*(*m),_2lda_ = 2*(*lda); \
  F77CALL (dgemv) (trans,&_2m_,n,alpha,(Real*)(a),&_2lda_, \
	x,incx,beta,(Real*)(y),incy) ; \
}
#endif

#endif

#endif
